// Tiny FAT16 implementation

#include "common.h"
#include "spi.h"
#include "fs.h"
#include "gps.h"

// debug
#include "lcd.h"
#include "tiny_printf.h"

/**************************************************************************
 * MM layer functions
 * mm_send_cmd()
 * mm_init()
 * mm_read_block()
 * mm_write_block()
 * spi_cs_assert()
 * spi_cs_deassert()
 **************************************************************************/

// MMC specific constants
#define MM_IDLE_TIMEOUT		1000
#define MM_CMD_TIMEOUT		1000
#define MM_BUSY_TIMEOUT		100000

#define MM_CMD0			0		// Reset card
#define MM_CMD1			1		// Ask if card reset has done
#define MM_CMD8			8
#define MM_CMD16			16		// Set block length for CMD17,24
#define MM_CMD17			17		// Single block read
#define MM_CMD24			24		// Single block write
#define MM_ACMD41			41
#define MM_CMD55			55

#define MM_R1_IDLE 		0x01
#define MM_R1_ERASE_RST		0x02
#define MM_R1_ILL_CMD		0x04
#define MM_R1_CRC_ERR		0x08
#define MM_R1_ERASE_SEQ_ERR 0x10
#define MM_R1_ADDR_ERR		0x20
#define MM_R1_PARAM_ERR		0x40
#define MM_RX_BUSFREE		0x80
#define MM_RX_OK			0x00
#define MM_RX_ERR			0xff


// Response type must be R1
static unsigned char mm_send_cmd(unsigned char cmd, unsigned char *arg)
{
	unsigned char rv;
	unsigned long i;

	spi_cs_assert();

	// Send command
	spi_trx(0x40 | (cmd & 0x3f));	// TX1 : "01" + CMD
	spi_trx(arg[3]);				// TX2 : arg31:24
	spi_trx(arg[2]);				// TX3 : arg23:16
	spi_trx(arg[1]);				// TX4 : arg15: 8
	spi_trx(arg[0]);				// TX5 : arg 7: 0
	if(cmd == MM_CMD8){
		spi_trx(0x87);			// TX6: CRC for CMD8 with 0x1aa
	} else {
		spi_trx(0x95);
	}							// TX6: CRC for CMD9 with 0x00
								// Note : CRC is not checked under SPI mode.
								// But required at very first CMD0 in MMC mode.
								// Use this hard-coded dummy CRC.
	// Receive response
	i = 0;
	do {
		rv = spi_trx(0xff);
		i++;
	} while((rv & 0x80) && (i < MM_CMD_TIMEOUT));

	if(i == MM_CMD_TIMEOUT){
		rv = MM_RX_ERR;
	}

	spi_cs_deassert();

	// Dummy clock
	// Some card require 8 dummy clocks with CS=H after each command.
	// See http://cwaweb.bai.ne.jp/~ohishi/zakki/sasicard.htm
	// I spent lots of time due to this problem.
	spi_trx(0xff);
	return rv;
}

unsigned char mm_init(void)
{
	unsigned long i;
	unsigned char rv;
	unsigned long dummy = 0;
	unsigned char data[4];

	spi_init();

	// Entering SPI mode
	spi_cs_deassert();

	// Send more than 74 clks
	for(i=0; i<10; i++){
		spi_trx(0xff);
	}

	// Reset card
	rv = mm_send_cmd(MM_CMD0, (unsigned char *)&dummy);
	if(rv & MM_RX_BUSFREE) return MM_RX_BUSFREE;
	if(~rv & MM_R1_IDLE) return MM_RX_ERR;

	// SDHC support (TBD)
	// Send CMD8
	// dummy = 0x1aa;
	// rv = mm_send_cmd(MM_CMD8, (unsigned char *)&dummy);

	// Try ACMD41 initialize first
	i = 0;
	do {
		rv = mm_send_cmd(MM_CMD55, (unsigned char *)&dummy);
		if(rv & MM_R1_ILL_CMD) goto mm_initialize;

		rv = mm_send_cmd(MM_ACMD41, (unsigned char *)&dummy);
		if(rv & MM_R1_ILL_CMD) goto mm_initialize;
		i++;
	} while ((rv & MM_R1_IDLE) && (i < MM_IDLE_TIMEOUT));

	if(~rv & MM_R1_IDLE) goto success;

	// MM initialize protocol
mm_initialize:		
	i = 0;
	do {
		rv = mm_send_cmd(MM_CMD1, (unsigned char *)&dummy);
		i++;
	} while ((rv & MM_R1_IDLE) && (i < MM_IDLE_TIMEOUT));

	if(~rv & MM_R1_IDLE) goto success;

	// Error	
	spi_cs_deassert();
	return MM_RX_ERR;

success:
	spi_cs_deassert();

	// Set to full speed (10Mbps)
	spi_fullspeed();
	return MM_RX_OK;
}

// Waiting the card return from busy state
unsigned char mm_sync_busy(void)
{
	unsigned char rv;
	unsigned long numwait =0;

	do {
		rv = spi_trx(0xff);
		numwait++;
	} while ((rv != 0xff) && (numwait < MM_BUSY_TIMEOUT));
	
	if(numwait == MM_BUSY_TIMEOUT){
		return 0;
	}
	return 1;
}

// Read single block
unsigned char mm_read_block(unsigned long addr, unsigned char *data)
{
	unsigned char rv;
	unsigned short i;

	rv = mm_send_cmd(MM_CMD17, (unsigned char *)&addr);
	if(rv){
		return rv;
	}

	spi_cs_assert();

	// Waiting for token.
	i = 0;
	do{
		rv = spi_trx(0xff);
		i++;
	}while((rv == 0xff) && (i <  MM_CMD_TIMEOUT));

	if(i == MM_CMD_TIMEOUT){
		spi_cs_deassert();
		return MM_RX_ERR;
	}

	for(i = 0; i < 512; i++){
		data[i] = spi_trx(0xff);
	}

	// Ignore CRC (2bytes)
	spi_trx(0xff);
	spi_trx(0xff);

	// Sync bysy
	mm_sync_busy();

	spi_cs_deassert();
	return MM_RX_OK;
}

// Write single block
unsigned char mm_write_block(unsigned long addr, unsigned char *tx_data)
{
	unsigned char rv;
	unsigned short i;

	rv = mm_send_cmd(MM_CMD24, (unsigned char *)&addr);
	if(rv){
		// Error
		return rv;
	}

	// Send following data
	spi_cs_assert();

	// The write command needs additiional 8 clocks before the block to be transffered.
	spi_trx(0xfe);

	for(i = 0; i < 512; i++){
		spi_trx(tx_data[i]);
	}

	// Needs ? CRC are issued by SD, some documents said.
	spi_trx(0xff);
	spi_trx(0xff);

	// Sync busy
	if(mm_sync_busy() == 0){
		spi_cs_deassert();
		return MM_RX_ERR;
	}
	spi_cs_deassert();
	// Success
	return MM_RX_OK;
}



/**************************************************************************
 * Ultra tiny FAT16 layer
 * fs_init()
 * fs_open()
 * fs_close()
 * fs_write()
 * fs_get_size()
 **************************************************************************/
#define b1toc(c) ((unsigned char)(c))

#define b4tol(c) \
 ( (((unsigned long)(*(c+3) & 0xff)) << 24) \
 | (((unsigned long)(*(c+2) & 0xff)) << 16) \
 | (((unsigned long)(*(c+1) & 0xff)) <<  8) \
 | (((unsigned long)(*(c+0) & 0xff)) <<  0) )


#define b2tol(c) \
 ( (((unsigned long)(*(c+1) & 0xff)) <<  8) \
 | (((unsigned long)(*(c+0) & 0xff)) <<  0) )

#define b1tol(c) ((unsigned long)(c))

#define b2tos(c) \
 ( (((unsigned short)(*(c+1) & 0xff)) << 8) \
 | (((unsigned short)(*(c+0) & 0xff)) << 0) )


#define ltob2(c,l) \
{ \
 *(c+1) = (unsigned char)((l >>  8)&0xff); \
 *(c+0) = (unsigned char)( l       &0xff); \
}

#define ltob4(c,l) \
{ \
 *(c+3) = (unsigned char)((l >> 24)&0xff); \
 *(c+2) = (unsigned char)((l >> 16)&0xff); \
 *(c+1) = (unsigned char)((l >>  8)&0xff); \
 *(c+0) = (unsigned char)( l       &0xff); \
}

#define stob2(c, s) \
{ \
 *(c+1) = (unsigned char)(s >> 8) & 0xff; \
 *(c+0) = (unsigned char) s       & 0xff; \
}
unsigned char mm_buff[512];
unsigned long topaddr = 0xffffffff;

#define MM_READ_BLOCK(rv, type, p, addr) \
{ \
	rv = MM_RX_OK; \
	if( (addr) < topaddr || (addr) >= topaddr + 512){ \
		topaddr = (addr) & 0xFFFFFE00; \
		(rv) = mm_read_block(topaddr, mm_buff); \
	} \
	(p) = (type *)(mm_buff + ((addr) - topaddr)%512); \
}

#define MM_WRITE_BLOCK(rv) \
{ \
	(rv) = mm_write_block(topaddr, mm_buff); \
}

typedef unsigned char Byte;

/* Data structure of Partition table on the MBR */
typedef struct {
	Byte bootDescriptor;
	Byte firstPartitionSector[3];
	Byte fileSystemDescriptor;
	Byte lastPartitionSector[3];
	Byte firstSectorNumbers[4];
	Byte numberOfSectors[4];
} PTBL_t;

#define OFFSET_TO_PTBL1 (446)		/* Offset to patrition table1 */

typedef struct {
	Byte    bytesPerSector[2];      /* bytes/sector */
	Byte    sectorsPerCluster;      /* sectors/cluster */
	Byte    reservedSectors[2];     /* reserved sector, beginning with sector 0 */
	Byte    numberOfFATs;           /* file allocation table */
	Byte    rootEntries[2];         /* root entry (512) */
	Byte    totalSectors[2];        /* partion total secter */
	Byte    mediaDescriptor;        /* 0xf8: Hard Disk */
	Byte    sectorsPerFAT[2];       /* sector/FAT (FAT32 always zero: see bigSectorsPerFAT) */
	Byte    sectorsPerTrack[2];     /* sector/track (not use) */
	Byte    heads[2];               /* heads number (not use) */
	Byte    hiddenSectors[4];       /* hidden sector number */
	Byte    bigTotalSectors[4];     /* total sector number */
} FAT16BPB_t;

#define OFFSET_TO_BPB (3+8)

/* Data structure of root directory entry */
typedef struct{
	Byte    name[8];				/* file name */
	Byte    extension[3];			/* file extention */
	Byte    attribute;
	Byte    reserved;
	Byte    createTimeMs;
	Byte    createTime[2];
	Byte    createDate[2];
	Byte    accessDate[2];
	Byte    clusterHighWord[2];
	Byte    updateTime[2];
	Byte    updateDate[2];
	Byte    cluster[2];
	Byte    fileSize[4];
} RDE_t;

/* FAT16 information */
typedef struct {
	unsigned long addr_fat1;			// Head address of FAT1
	unsigned long addr_fat2;			// Head address of FAT2
	unsigned long addr_rde;			// Head address of RootDirEntry
	unsigned long addr_uda;			// Head address of UserDataArea
	unsigned long num_uda_cluster;		// Number of cluster of UserDataArea
	unsigned char sectors_per_cluster;	// Number of sectors per cluster
	unsigned short sectors_per_fat;	// Number of sectors per FAT
	unsigned short num_rde;			// Number of root entries
} FAT16Info_t;

FAT16Info_t fat16info;					// We support only 1 FAT16 concurrently.

typedef struct {
	unsigned long  addr_rde;			// RDE address of this file
	unsigned long  curr_cluster;		// Current cluster
	unsigned long  curr_sector;		// Current sector
	unsigned long  curr_size;			// Current file size
	unsigned short buff_len;			// Current data length in the buffer
	unsigned char  buff[512];			// Buffer
} OPENFILEINFO_t;

OPENFILEINFO_t ofinfo[1];				// Currently supports only 1 open file concurrently

// Initialize FAT system
unsigned char fs_init(void)
{
	PTBL_t *ptbl;
	FAT16BPB_t *fat16bpb;
	unsigned long addr_pbr1;
	unsigned char rv;

	// Init mm layer
	rv = mm_init();
	switch(rv){
	case MM_RX_OK:
		break;
	case MM_RX_BUSFREE:
		return FS_NOCARD;
	default:
		return FS_IOERR;
	}

	// Read Partition table 1 (on MBR)
	MM_READ_BLOCK(rv, PTBL_t, ptbl, 0 + OFFSET_TO_PTBL1);
	if(rv) return FS_IOERR;

	// Check filesystem type
	// 1:FAT12, 4:FAT16(less than 32MB), 5:Ext DOS partition,
	// 6:FAT16(more than 32MB), 0xb:FAT32(more than 2GB),
	// 0xc:FAT32 Int32h ext, 0xe:FAT16 Int32h ext,
	// 0xf:Ext DOS partition Int32h ext
	if(ptbl->fileSystemDescriptor != 0x04 &&
	   ptbl->fileSystemDescriptor != 0x06 ){
		return FS_NOTFAT16;
	}

	// Get offset to PBR1
	addr_pbr1 = b4tol(ptbl->firstSectorNumbers);
	addr_pbr1 <<= 9;

	// Read PBR and parse FAT16 information
	MM_READ_BLOCK(rv, FAT16BPB_t, fat16bpb, addr_pbr1 + OFFSET_TO_BPB);
	if(rv) return FS_IOERR;

	fat16info.addr_fat1 = addr_pbr1 + (b2tol(fat16bpb->reservedSectors) << 9);
	fat16info.addr_fat2 = fat16info.addr_fat1 + (b2tol(fat16bpb->sectorsPerFAT) << 9);

	fat16info.addr_rde  = fat16info.addr_fat2 + (b2tol(fat16bpb->sectorsPerFAT) << 9);
	fat16info.addr_uda  = fat16info.addr_rde  + (b2tol(fat16bpb->rootEntries) << 5);

	fat16info.sectors_per_cluster = b1toc(fat16bpb->sectorsPerCluster);
	fat16info.sectors_per_fat = b2tos(fat16bpb->sectorsPerFAT);

	fat16info.num_rde = b2tos(fat16bpb->rootEntries);
	fat16info.num_uda_cluster = 
		( b4tol(fat16bpb->bigTotalSectors) - b2tol(fat16bpb->reservedSectors)
		- b1tol(fat16bpb->numberOfFATs) * b2tol(fat16bpb->sectorsPerFAT)
		- b2tol(fat16bpb->rootEntries) * 32 / b2tol(fat16bpb->bytesPerSector) )
		/ fat16info.sectors_per_cluster;

	return FS_OK;
}

#define ATTR_ARCHIVE	0x20
#define ATTR_DIRECTORY	0x10
#define ATTR_VOLUME	0x08
#define ATTR_HIDDEN	0x04
#define ATTR_SYSTEM	0x02
#define ATTR_READONLY	0x01

#define toupper(x)		((x) > 0x60 ? (x) - 0x20 : (x))

// Open file
unsigned char fs_open(const char *filename, unsigned char mode)
{
	RDE_t *rde;
	unsigned long i, j;
	unsigned short *cluster, daytime;
	unsigned char rv;
	unsigned long addr;

	// Looking for unused root directory entry
	for(i = 0; i < fat16info.num_rde; i++){
		ofinfo[0].addr_rde = fat16info.addr_rde + sizeof(RDE_t) * i;
		MM_READ_BLOCK(rv, RDE_t, rde, ofinfo[0].addr_rde);
		if(rv) return FS_IOERR;

		if(mode == FS_MODE_WR){
			if(rde->name[0] == 0x00 || rde->name[0] == 0x05 || rde->name[0] == 0xe5){
				// Find unused (or deleted) entry
				break;
			}
		}
	}
	if(i == fat16info.num_rde){
		// No room for new file
		return FS_FULL;
	}

	// Looking for unused cluster from FAT1
	for(ofinfo[0].curr_cluster = 2; ofinfo[0].curr_cluster < fat16info.num_uda_cluster; ofinfo[0].curr_cluster++){
		addr = fat16info.addr_fat1 + 2 * ofinfo[0].curr_cluster;
		MM_READ_BLOCK(rv, unsigned short, cluster, addr);
		if(rv) return FS_IOERR;
		if(*cluster == 0){
			// Find unused cluster
			// Touch this cluster as the last cluster of this file.
			//*cluster = 0xffff;
			stob2((char *)cluster, 0xffff);
			MM_WRITE_BLOCK(rv);
			if(rv) return FS_IOERR;
			break;
		}
	}
	if(ofinfo[0].curr_cluster == fat16info.num_uda_cluster){
		// No room for new file
		return FS_FULL;
	}

	// Update secondary FAT
	addr = fat16info.addr_fat2 + 2 * ofinfo[0].curr_cluster;
	MM_READ_BLOCK(rv, unsigned short, cluster, addr);
	if(rv) return FS_IOERR;
	//*cluster = 0xffff;
	stob2((char *)cluster, 0xffff);
	MM_WRITE_BLOCK(rv);
	if(rv) return FS_IOERR;

	// Update RDE
	MM_READ_BLOCK(rv, RDE_t, rde, ofinfo[0].addr_rde);
	if(rv) return FS_IOERR;

	// Copy filename
	for(i = 0; i< 8 && *filename && *filename != '.'; i++){
		rde->name[i] = toupper(*filename);
		filename++;
	}
	for( ; i<8; i++){
		rde->name[i] = 0x20;
	}
	if(*filename == '.') filename++;

	for(i=0; i<3 && *filename; i++){
		rde->extension[i] = toupper(*filename);
		filename++;
	}
	for( ; i<3; i++){
		rde->extension[i] = 0x20;
	}

	// Update attribute
	rde->attribute = ATTR_ARCHIVE;

	// Update Time and Date
	daytime =  ((unsigned short)(gps_info.daytime.hour & 0x1f) << 11)
		    | ((unsigned short)(gps_info.daytime.min & 0x3f) << 5)
		    | ((unsigned short)(gps_info.daytime.sec >> 1) & 0x1f);
	stob2(rde->updateTime, daytime);

	daytime = (((unsigned short)(gps_info.daytime.year +20) & 0x7f) << 9)
		    | ((unsigned short)(gps_info.daytime.mon & 0x0f) << 5)
		    | (unsigned short)(gps_info.daytime.day & 0x1f);
	stob2(rde->updateDate, daytime);

	// Update start cluster number
	ltob2(rde->cluster, ofinfo[0].curr_cluster);

	// Upddate file size
	ltob4(rde->fileSize, 0);

	// Write RDE
	MM_WRITE_BLOCK(rv);
	if(rv) return FS_IOERR;

	ofinfo[0].curr_sector = 0;
	ofinfo[0].curr_size = 0;
	ofinfo[0].buff_len = 0;

	return FS_OK;
}

// Close file
void fs_close(void)
{
	unsigned long addr;
	unsigned char rv;
	unsigned short daytime;
	RDE_t *rde;

	// Purge buffered data
	addr = fat16info.addr_uda
		 + (ofinfo[0].curr_cluster -2) * ((unsigned long)fat16info.sectors_per_cluster << 9)
		 + (ofinfo[0].curr_sector << 9);
	mm_write_block(addr, ofinfo[0].buff);

	// Update RDE
	MM_READ_BLOCK(rv, RDE_t, rde, ofinfo[0].addr_rde);

	// Update Time and Date
	daytime =  ((unsigned short)(gps_info.daytime.hour & 0x1f) << 11)
		    | ((unsigned short)(gps_info.daytime.min & 0x3f) << 5)
		    | ((unsigned short)(gps_info.daytime.sec >> 1) & 0x1f);
	stob2(rde->updateTime, daytime);

	daytime = (((unsigned short)(gps_info.daytime.year +20) & 0x7f) << 9)
		    | ((unsigned short)(gps_info.daytime.mon & 0x0f) << 5)
		    | (unsigned short)(gps_info.daytime.day & 0x1f);
	stob2(rde->updateDate, daytime);

	// Update filesize
	ltob4(rde->fileSize, ofinfo[0].curr_size);
	MM_WRITE_BLOCK(rv);
}

// Write data to file
unsigned char fs_write(const char *data, long data_len)
{
	unsigned char rv;
	unsigned long len, data_ofs=0, addr;
	unsigned short next_cluster, *cluster, daytime;
	RDE_t *rde;

	while(data_len){
		len = MIN(512 - ofinfo[0].buff_len, data_len);
		memcpy(ofinfo[0].buff + ofinfo[0].buff_len, data + data_ofs, len);
		ofinfo[0].buff_len += len;
		data_ofs += len;
		data_len -= len;
		ofinfo[0].curr_size += len;
		if(ofinfo[0].buff_len == 512){
			ofinfo[0].buff_len = 0;
			// Purge the sector
			addr = fat16info.addr_uda
				 + (ofinfo[0].curr_cluster -2) * ((unsigned long)fat16info.sectors_per_cluster << 9)
				 + (ofinfo[0].curr_sector << 9);
			rv = mm_write_block(addr, ofinfo[0].buff);
			if(rv) return FS_IOERR;
			
			if(ofinfo[0].curr_sector == fat16info.sectors_per_cluster -1){
				// Allocate next cluster
				for(next_cluster = ofinfo[0].curr_cluster +1; ; next_cluster ++){
					// Look for unused cluster
					addr = fat16info.addr_fat1 + 2 * next_cluster;
					MM_READ_BLOCK(rv, unsigned short, cluster, addr);
					if(rv) return FS_IOERR;
					if(*cluster == 0){
						// Find unused cluster
						//*cluster = 0xffff;	// mark it used and the last
						stob2((char *)cluster, 0xffff);
						MM_WRITE_BLOCK(rv);
						if(rv) return FS_IOERR;
						break;
					}
				}
				
				// Update curr cluster
				addr = fat16info.addr_fat1 + 2 * ofinfo[0].curr_cluster;
				MM_READ_BLOCK(rv, unsigned short, cluster, addr);
				if(rv) return FS_IOERR;
				stob2((char *)cluster, next_cluster);
				// *cluster = next_cluster;
				MM_WRITE_BLOCK(rv);
				if(rv) return FS_IOERR;
				
				// Update secondary FAT
				addr = fat16info.addr_fat2 + 2 * next_cluster;
				MM_READ_BLOCK(rv, unsigned short, cluster, addr);
				if(rv) return FS_IOERR;
				//*cluster = 0xffff;
				stob2((char *)cluster, 0xffff);
				MM_WRITE_BLOCK(rv);
				if(rv) return FS_IOERR;

				addr = fat16info.addr_fat2 + 2 * ofinfo[0].curr_cluster;
				MM_READ_BLOCK(rv, unsigned short, cluster, addr);
				if(rv) return FS_IOERR;
				//*cluster = next_cluster;
				stob2((char *)cluster, next_cluster);
				MM_WRITE_BLOCK(rv);
				if(rv) return FS_IOERR;

				ofinfo[0].curr_cluster = next_cluster;
				ofinfo[0].curr_sector = 0;

			} else {
				// Next sector
				ofinfo[0].curr_sector ++;
			}
		}
	}

	// Update RDE
	MM_READ_BLOCK(rv, RDE_t, rde, ofinfo[0].addr_rde);
	if(rv) return FS_IOERR;

	// Update Time and Date
	daytime = ((unsigned short)(gps_info.daytime.hour & 0x1f) << 11)
		    | ((unsigned short)(gps_info.daytime.min & 0x3f) << 5)
		    | ((unsigned short)(gps_info.daytime.sec >> 1) & 0x1f);
	stob2(rde->updateTime, daytime);

	daytime = (((unsigned short)(gps_info.daytime.year +20) & 0x7f) << 9)
		    | ((unsigned short)(gps_info.daytime.mon & 0x0f) << 5)
		    | (unsigned short)(gps_info.daytime.day & 0x1f);
	stob2(rde->updateDate, daytime);

	// Update file size
	ltob4(rde->fileSize, ofinfo[0].curr_size);

	MM_WRITE_BLOCK(rv);
	if(rv) return FS_IOERR;

	return FS_OK;
}

// Retrieve current file size
unsigned long fs_get_size(void)
{
	return ofinfo[0].curr_size;
}
