/********************************************
* Software serial communication driver
* First written on 09/07/25 H.Momose
*********************************************/
#include <htc.h>
#include "serial.h"

/* serial timing consideration
 *  clock (use internal clock 40MHz)
 *  baud rate 9600 (bit time 1.042E-4 s)
 */  

/* Serial connection
 * TXD RB4 (not implemented)
 * RXD RA4/TOCKI
 */

unsigned char ser_state;
unsigned char ser_rxdata;
unsigned char ser_txdata;
unsigned char ser_bitpos;

// CPU operates @ 40MHz
#define TMRV_STBIT		 (61)	// 256-195
#define TMRV_BIT		(131)	// 256-125

#define ser_rxwaiting() \
{ \
	ser_state = SER_RXWAITING;	\ // Set state				
	T0CS   = 1;				\ // Select RA4 as TMR0 clock	
	PSA    = 1;				\ // No prescaler			
 	TMR0L  = 0xff;			\ // Set TMR0 at full count
	TMR0IF = 0;				\
	TMR0IE = 1;				\ // Enable TMR0 INT
}

// Serial interrupt
void ser_isr(void)
{
	TMR0IF = 0;
	
	switch(ser_state){

	case SER_RXWAITING:
		TMR0L = TMRV_STBIT;
		PSA  = 0; // Use prescaler		
		T0CS = 0; // Use clock for TMR0
		ser_state = SER_RXBUSY;
		ser_bitpos = 0;
		ser_rxdata = 0;
		break;

	case SER_RXBUSY:		
		if(ser_bitpos == 8){
			// STOP bit
			if(RA4){
				// Must be 1 (w/o error)
				ser_state = SER_RXREADY;
			} else {
				// error
				ser_state = SER_RXERR;
			}
			TMR0IE = 0;
			break;
		}
		ser_rxdata >>= 1;
		// Data in progress
		if(RA4){
			// Data comes with LSB first.
			ser_rxdata |= 0x80;
		}
		ser_bitpos ++;
		TMR0L += TMRV_BIT;
		break;

	case SER_TXBUSY:
		if(ser_bitpos == 9){
			// TX done
			ser_rxwaiting();
			break;
		} else if (ser_bitpos == 8){
			// Stop bit
			RB4 = 1;
		} else {
			// Sending data
			if(ser_txdata & 0x01){
				RB4 = 1;
			} else {
				RB4 = 0;
			}
			ser_txdata >>= 1;
		}
		ser_bitpos ++;
		TMR0L += TMRV_BIT;
		break;

	default:
		ser_state = SER_RXERR;
		TMR0IE = 0;
		break;
	}
}

void ser_init(void)
{
	// Set port direction
	TRISB4 = 0;		// RB4 output (TXD)
	TRISA4 = 1;		// RA4 input  (RXD)
	RB4 = 1;

	// Followings are on same register (T0CON)
	TMR0ON = 1;	// Enable TMR0
	T08BIT = 1;	// Use 8bit counter instead of 16
	T0SE = 1;		// Invert RA4 for TMR0 clock
	T0PS2 = 0;	// Prescaler select 1/8 
	T0PS1 = 1;
	T0PS0 = 0;

	// Entering RXWAITING state
	ser_rxwaiting();
}

char ser_getchar(void)
{
	// Copy received data to buffer
	char data = (char)ser_rxdata;

	// Entering RXWAITING state for next data reception
	ser_rxwaiting();
	return data;
}

// Note: RX inprogress data may be destroyed with this function
void ser_putchar(char data)
{
	while( ser_getstate() == SER_RXBUSY || ser_getstate() == SER_TXBUSY) continue;
	TMR0IE = 0;				// Disable TMR0 INT
	TMR0L = TMRV_BIT;			// Set Start bit duration
	RB4 = 0;
	PSA = 0;					// Use prescaler
	T0CS = 0;					// Use Fosc/4 as clock
	ser_txdata = (unsigned char)data;			// Set txdata to buffer
	ser_state = SER_TXBUSY;
	ser_bitpos = 0;	
	TMR0IF = 0;				// Clear TMR0 Int flag
	TMR0IE = 1;
}
